#include "connection.h"
#include "config.h"
#include "format.h"
#include "log.h"
#include "match.h"
#include "rules.h"

#include <sys/types.h>	/* connect */
#include <sys/socket.h>	/* connect */
#include <arpa/inet.h>	/* ntohs, inet_pton */
#include <string.h>	/* memcmp, strchr */



static int
match_inet(const struct connection_rule *rule,
		const struct sockaddr *from_addr, socklen_t from_addrlen,
		const struct sockaddr *to_addr, socklen_t to_addrlen)
{
	char rule_str[ADDR_LEN] = "";
	char to_addr_str[ADDR_LEN] = "-";
	char from_addr_str[ADDR_LEN] = "-";
	int match = 0;
	uint16_t port;

	config_rule_to_str(rule_str, ADDR_LEN, rule);
	sockaddr_to_str(from_addr_str, ADDR_LEN, from_addr, from_addrlen);
	sockaddr_to_str(to_addr_str, ADDR_LEN, to_addr, to_addrlen);
	log_debug("rule: (%s) -- conn: (from=\"%s\" to=\"%s\")",
			rule_str, from_addr_str, to_addr_str);

	log_debug("~~ from: rule=%s/%d this=%s (%p)",
			rule->from_addr ? : "-", rule->from_port,
			from_addr_str, (void *) from_addr);
	log_debug("~~   to: rule=%s/%d this=%s (%p)",
			rule->to_addr ? : "-", rule->to_port,
			to_addr_str, (void *) to_addr);

	if (rule->from_addr && from_addr) {
		log_debug("to:: rule=%s this=%s", rule->from_addr, from_addr_str);
		if (match_addr(rule->from_addr, from_addr_str)) {
			printf("MATCH!!\n");
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->from_port && from_addr) {
		port = ntohs(((struct sockaddr_in *) from_addr)->sin_port);
		log_debug("to:: rule=%i this=%u", rule->from_port, port);
		if (rule->from_port == port) {
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->to_addr && to_addr) {
		log_debug("to:: rule=%s this=%s", rule->to_addr, to_addr_str);
		if (to_addr && match_addr(rule->to_addr, to_addr_str)) {
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->to_port && to_addr) {
		port = ntohs(((struct sockaddr_in *) to_addr)->sin_port);
		log_debug("to:: rule=%i this=%u", rule->to_port, port);
		if (rule->to_port == port) {
			match = 1;
		} else {
			return 0;
		}
	}
	return match;
}



static int
match_inet6(const struct connection_rule *rule,
		const struct sockaddr *from_addr, socklen_t from_addrlen,
		const struct sockaddr *to_addr, socklen_t to_addrlen)
{
	char rule_str[ADDR_LEN] = "";
	char to_addr_str[ADDR_LEN] = "-";
	char from_addr_str[ADDR_LEN] = "-";
	int match = 0;
	uint16_t port;

	config_rule_to_str(rule_str, ADDR_LEN, rule);
	sockaddr_to_str(to_addr_str, ADDR_LEN, to_addr, to_addrlen);
	sockaddr_to_str(from_addr_str, ADDR_LEN, from_addr, from_addrlen);
	log_debug("rule: (%s) -- conn: (from=\"%s\" to=\"%s\")",
			rule_str, to_addr_str, from_addr_str);

	if (rule->from_addr && from_addr) {
		if (match_addr(rule->from_addr, from_addr_str)) {
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->from_port && from_addr) {
		port = ntohs(((struct sockaddr_in6 *) from_addr)->sin6_port);
		if (rule->from_port == port) {
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->to_addr && to_addr) {
		if (to_addr && match_addr(rule->to_addr, to_addr_str)) {
			match = 1;
		} else {
			return 0;
		}
	}
	if (rule->to_port && to_addr) {
		port = ntohs(((struct sockaddr_in6 *) to_addr)->sin6_port);
		if (rule->to_port == port) {
			match = 1;
		} else {
			return 0;
		}
	}
	return match;
}



static int
match_single(const struct connection_rule *rule,
		const struct sockaddr *from_addr, socklen_t from_addrlen,
		const struct sockaddr *to_addr, socklen_t to_addrlen)
{
	const struct sockaddr *addr;
	addr = to_addr ? : from_addr;
	switch (addr->sa_family) {
		case AF_INET:
			return match_inet(rule, from_addr, from_addrlen,
					to_addr, to_addrlen);
		case AF_INET6:
			return match_inet6(rule, from_addr, from_addrlen,
					to_addr, to_addrlen);
		default:
			return 0;
	}
}



static void
write_spoofed_sockaddr(struct sockaddr *addr, socklen_t *addrlen,
		struct connection_rule *rule)
{
	if (strchr(rule->spoof_addr, ':')) {
		struct sockaddr_in6 *inet;
		inet = (struct sockaddr_in6 *) addr;
		inet_pton(AF_INET6, rule->spoof_addr, &(inet->sin6_addr));
		inet->sin6_port = htons(rule->spoof_port);
	} else {
		struct sockaddr_in *inet;
		inet = (struct sockaddr_in *) addr;
		inet_pton(AF_INET, rule->spoof_addr, &(inet->sin_addr));
		inet->sin_port = htons(rule->spoof_port);
	}
}



/* FIXME for some reason gcc doesn't like to have "const struct connection_rule ** */
static int
filter_connection(struct connection_rule **rules,
		const struct sockaddr *from_addr, socklen_t from_addrlen,
		const struct sockaddr *to_addr, socklen_t to_addrlen,
		struct sockaddr *spoof_addr, socklen_t *spoof_addrlen,
		struct action *policy, struct action *action)
{
	struct connection_rule **rule;

	{
		char from[ADDR_LEN] = "-";
		char to[ADDR_LEN] = "-";
		sockaddr_to_str(from, ADDR_LEN, from_addr, from_addrlen);
		sockaddr_to_str(to, ADDR_LEN, to_addr, to_addrlen);
		log_debug("filter: from=\"%s\" to=\"%s\"", from, to);
	}

	for (rule = rules; *rule; rule++) {
		if (match_single(*rule, from_addr, from_addrlen,
					to_addr, to_addrlen)) {
			if (spoof_addr && (*rule)->spoof_addr) {
				write_spoofed_sockaddr(spoof_addr,
						spoof_addrlen,
						*rule);
			}
			action_copy(action, &((*rule)->policy));
			return 1;
		}
	}

	action_copy(action, policy);

	return 0;
}



int
filter_outbound(const struct sockaddr *addr, socklen_t addrlen,
		struct sockaddr *spoof_addr, socklen_t *spoof_addrlen,
		struct action *action)
{
	action->set_ret = action->set_err = 0;

	if (spoof_addr) {
		memcpy(spoof_addr, addr, addrlen);
		*spoof_addrlen = addrlen;
	}
	log_debug("OUTPUT");

	if (condnet_config) {
		return filter_connection(condnet_config->out,
				NULL, 0, addr, addrlen,
				spoof_addr, spoof_addrlen,
				&(condnet_config->out_policy),
				action);
	}
	return 0;
}



int
filter_inbound(const struct sockaddr *addr, socklen_t addrlen,
		struct sockaddr *spoof_addr, socklen_t *spoof_addrlen,
		struct action *action)
{
	action->set_ret = action->set_err = 0;

	if (spoof_addr) {
		memcpy(spoof_addr, addr, addrlen);
		*spoof_addrlen = addrlen;
	}
	log_debug("INPUT");

	config_setup();

	if (condnet_config) {
		return filter_connection(condnet_config->in,
				addr, addrlen, NULL, 0,
				spoof_addr, spoof_addrlen,
				&(condnet_config->in_policy),
				action);
	}
	return 0;
}
